#define _LARGEFILE64_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>
#include <stdarg.h>

#include "sd.h"
#include "utils.h"


typedef int (*unlink_t)	(const char *pathname);
typedef int (*truncate_t) (const char *path, off_t length);
typedef int (*ftruncate_t) (int fd, off_t length);
typedef int (*truncate64_t) (const char *path, off64_t length);
typedef int (*ftruncate64_t) (int fd, off64_t length);
typedef int (*open_t) (const char *pathname, int flags, ...);
typedef int (*creat_t) (const char *pathname, mode_t mode);
typedef int (*open64_t) (const char *pathname, int flags, ...);
typedef int (*creat64_t) (const char *pathname, mode_t mode);

static unlink_t unlink_org_func = NULL;
static truncate_t truncate_org_func = NULL;
static ftruncate_t ftruncate_org_func = NULL;
static truncate64_t truncate64_org_func = NULL;
static ftruncate64_t ftruncate64_org_func = NULL;
static open_t open_org_func = NULL;
static open64_t open64_org_func = NULL;


void get_pointers(void)
{
	static char have_pointers = 0;

	if (!have_pointers)
	{
		unlink_org_func = (unlink_t)retrieve_lib_function_pointer("unlink");
		truncate_org_func = (truncate_t)retrieve_lib_function_pointer("truncate");
		ftruncate_org_func = (ftruncate_t)retrieve_lib_function_pointer("ftruncate");
		truncate64_org_func = (truncate64_t)retrieve_lib_function_pointer("truncate64");
		ftruncate64_org_func = (ftruncate64_t)retrieve_lib_function_pointer("ftruncate64");
		open_org_func = (open_t)retrieve_lib_function_pointer("open");
		open64_org_func = (open64_t)retrieve_lib_function_pointer("open64");

		have_pointers = 1;
	}
}

int unlink(const char *pathname)
{
	syslog(LOG_DEBUG, "libsd: unlink(%s)", pathname);

	get_pointers();

	/* secure delete */
	if (secure_erase_fname(pathname, (off64_t)0) != 0)
	{
		syslog(LOG_ERR, "libsd: unlink failed");
		return -1;
	}

	return unlink_org_func(pathname);
}

int truncate(const char *path, off_t length)
{
	syslog(LOG_DEBUG, "libsd: truncate(%s, %ld)", path, length);

	get_pointers();

	/* secure truncate */
	if (secure_erase_fname(path, (off64_t)length) != 0)
	{
		syslog(LOG_ERR, "libsd: truncate failed");
		return -1;
	}

	return truncate_org_func(path, length);
}

int ftruncate(int fd, off_t length)
{
	get_pointers();

	/* secure truncate */
	if (secure_erase_fd(fd, (off64_t)length) != 0)
	{
		syslog(LOG_ERR, "libsd: ftruncate failed");
		return -1;
	}

	return ftruncate_org_func(fd, length);
}

int truncate64(const char *path, off64_t length)
{
	syslog(LOG_DEBUG, "libsd: truncate64(%s, %lld)", path, length);

	get_pointers();

	/* secure truncate */
	if (secure_erase_fname(path, length) != 0)
	{
		syslog(LOG_ERR, "libsd: truncate64 failed");
		return -1;
	}

	return truncate64_org_func(path, length);
}

int ftruncate64(int fd, off64_t length)
{
	get_pointers();

	/* secure truncate */
	if (secure_erase_fd(fd, length) != 0)
	{
		syslog(LOG_ERR, "libsd: ftruncate64 failed");
		return -1;
	}

	return ftruncate64_org_func(fd, length);
}

int open(const char *pathname, int flags, ...)
{
        va_list ap;
        mode_t mode;

	syslog(LOG_DEBUG, "libsd: open(%s, %d)", pathname, flags);

	get_pointers();

        va_start(ap, flags);
        mode = va_arg(ap, mode_t);
        va_end(ap);

	/* truncate new file? */
	if (flags & O_TRUNC)
	{
		int rc = file_exists(pathname);

		/* error while checking for file? */
		if (rc == -1)
		{
			syslog(LOG_ERR, "libsd: open failed");
			return -1;
		}

		/* file did exist? then delete it */
		if (rc == 1)
		{
			/* then secure delete it first */
			if (secure_erase_fname(pathname, 0) != 0)
			{
				syslog(LOG_ERR, "libsd: open failed");
				return -1;
			}
		}
	}

	return open_org_func(pathname, flags, mode);
}

int creat(const char *pathname, mode_t mode)
{
	syslog(LOG_DEBUG, "libsd: creat(%s, %x)\n", pathname, mode);

	get_pointers();

	return open(pathname, O_CREAT|O_WRONLY|O_TRUNC, mode);
}

int open64(const char *pathname, int flags, ...)
{
        va_list ap;
        mode_t mode;

	syslog(LOG_DEBUG, "libsd: open64(%s, %d)", pathname, flags);

	get_pointers();

        va_start(ap, flags);
        mode = va_arg(ap, mode_t);
        va_end(ap);

	/* truncate new file? */
	if (flags & O_TRUNC)
	{
		int rc = file_exists(pathname);

		/* error while checking for file? */
		if (rc == -1)
		{
			syslog(LOG_ERR, "libsd: open64 failed");
			return -1;
		}

		/* file did exist? then delete it */
		if (rc == 1)
		{
			/* then secure delete it first */
			if (secure_erase_fname(pathname, 0) != 0)
			{
				syslog(LOG_ERR, "libsd: open64 failed");
				return -1;
			}
		}
	}

	return open64_org_func(pathname, flags, mode);
}

int creat64(const char *pathname, mode_t mode)
{
	syslog(LOG_DEBUG, "libsd: creat64(%s, %x)\n", pathname, mode);

	get_pointers();

	return open64(pathname, O_CREAT|O_WRONLY|O_TRUNC, mode);
}
