#define _LARGEFILE64_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>

#include "utils.h"

#define BLOCK_SIZE 512
#define ERASE_0x00	0	/* 0x00 */
#define ERASE_0xff	1	/* 0xff */
#define ERASE_RANDOM	2	/* random values */

int do_secure_erase(int fd, off64_t new_size, off64_t org_size)
{
	char err = 0;
	int loop;

	/* init. random library */
	srand48(get_seed());

	/* go through all methods */
	for(loop=0; loop<3; loop++)
	{
		off64_t offset;
		char buffer[BLOCK_SIZE];

		/* seek the first byte after current position */
		if (lseek64(fd, new_size, SEEK_SET) == -1)
		{
			syslog(LOG_ERR, "libsd: cannot seek in file to position %lld (%m)", new_size);
			err = -1;
			break;
		}

		if (loop == ERASE_0x00)					/* 0x00 */
			memset(buffer, 0x00, sizeof(buffer));
		else if (loop == ERASE_0xff)				/* 0xff */
			memset(buffer, 0xff, sizeof(buffer));

		/* overwrite original file */
		for(offset=new_size; offset<org_size; offset+=BLOCK_SIZE)
		{
			int cur_block_size = min(org_size - offset, BLOCK_SIZE);

			if (loop == ERASE_RANDOM)			/* random values */
			{
				int buffer_offset;

				for(buffer_offset=0; buffer_offset<cur_block_size; buffer_offset++)
					buffer[buffer_offset] = (char)limited_lrand48(255);
			}

			if (WRITE(fd, buffer, cur_block_size) != cur_block_size)
			{
				syslog(LOG_ERR, "libsd: error writing %d bytes to file at offset %lld (%m)", cur_block_size, offset);
				err = -1;
				break;	/* only break for this iteration: try continuing with other
					 * erase methods
					 */
			}
		}

		/* force file to disk */
		sync();
	}

	return 0;
}

int secure_erase_fd(int fd, off64_t new_size)
{
	off64_t org_size;
	struct stat64 file_info;

	/* determine size of file */
	if (fstat64(fd, &file_info) == -1)
	{
		syslog(LOG_ERR, "libsd: problem determing size of file (%m)");
	}
	org_size = file_info.st_size;

	/* only operate on files, not on devices/pipes/etc. */
	if (!S_ISREG(file_info.st_mode))
	{
		syslog(LOG_DEBUG, "libsd: not truncating file: not a regular file");
		return 0;
	}

	/* only do if file became smaller */
	if (new_size >= org_size)
	{
		syslog(LOG_DEBUG, "libsd: not truncating file (org size: %lld, new size: %lld)", org_size, new_size);
		return 0;
	}

	syslog(LOG_DEBUG, "libsd: truncating file from %lld to %lld bytes", org_size, new_size);

	return do_secure_erase(fd, new_size, org_size);
}

int secure_erase_fname(const char *file_name, off64_t new_size)
{
	int rc;
	int fd;
	off64_t org_size;
	struct stat64 file_info;

	/* determine size of file */
	if (stat64(file_name, &file_info) == -1)
	{
		syslog(LOG_ERR, "libsd: problem determing size of file %s (%m)", file_name);
		return -1;
	}
	org_size = file_info.st_size;

	/* only operate on files, not on devices/pipes/etc. */
	if (!S_ISREG(file_info.st_mode))
	{
		syslog(LOG_DEBUG, "libsd: not truncating file %s: not a regular file", file_name);
		return 0;
	}

	/* only do if file became smaller */
	if (new_size >= org_size)
	{
		syslog(LOG_DEBUG, "libsd: not truncating file %s (org size: %lld, new size: %lld)", file_name, org_size, new_size);
		return 0;
	}

	syslog(LOG_DEBUG, "libsd: truncating file %s from %lld to %lld bytes", file_name, org_size, new_size);

	/* open the file */
	/* this will get the open of this lib, but since we're not
	 * doing any truncates here, that is safe
	 */
	fd = open64(file_name, O_WRONLY | O_SYNC);
	if (fd == -1)
	{
		syslog(LOG_ERR, "libsd: problem opening %s for writing (%m)", file_name);
		return -1;
	}

	rc = do_secure_erase(fd, new_size, org_size);

	close(fd);

	return rc;
}
